//---------------------------------------------------------------------------
// File: TCP_Dummy_Sink_for_AudioStreams.c
// Date: 2012-04-30
// Author:  Wolfgang Buescher, DL4YHF
// Purpose: Simple TCP client; replaces a 'real' audio stream client (sink)
//          like Paul Nicholson's Live VLF Natural Radio server (http://abelian.org/vlf/) .
//          THIS module (TCP_Dummy_Sink_for_AudioStreams.c) doesn't care for
//          the data which it actually receives; it doesn't try to decode
//          the received Ogg/Vorbis stream but it can save the input
//          in a normal disk file.
//     Main purpose: Test for Spectrum lab. In the output stream configuration,
//          instead of  server = rawtcp://67.***.***.**:78XX
//          use         server = rawtcp://127.0.0.1:12345
//          to send the Ogg/Vorbis stream to this 'dummy sink' running on
//          the same PC as Spectrum Lab (IP = localhost, 1234 = listening port)
//---------------------------------------------------------------------------

// #include "winsock.h"
#include "winsock2.h"
#include "stdio.h"
#include "conio.h"
#pragma hdrstop

//---------------------------------------------------------------------------

char acReadBuffer[1024*1024]; // 2012-04-26: Increased to 1 MByte (!) to see how large the chunks may actually get


//---------------------------------------------------------------------------
void wait_n_seconds_with_indicator( int nSeconds )
{ // Waits for the specified number of seconds, but shows we're still alive
  int i;
  printf("\r\n Sleeping for %d seconds ..    ", (int)nSeconds);
  for(i=0; i<nSeconds; ++i)
   { printf("\x08\x08%02d", i );
     Sleep(1000); // argument in milliseconds
   }
  printf(" ..back at work");

}

//---------------------------------------------------------------------------
void show_help( void )
{
  printf("\r\nDummy 'test sink' for TCP/IP streams. Commands:");
  printf("\r\n H : show help");
  printf("\r\n p : pause for 10 seconds to check the peer's output buffering");
  printf("\r\n P : Pause for 30 seconds to really torture the peer");
  printf("\r\n Q : quit (use ctrl-c when stuck in recv)");
  printf("\r\n R : reset the current connection, and listen for new connection");
  printf("\r\n");
}


//---------------------------------------------------------------------------
#pragma argsused
int main(int argc, char* argv[])
{
  char *pcAddress = "127.0.0.1"; // use "localhost" for simplicity
  int   nPort     = 12345;
  unsigned long nInterfaceAddr;
  int   iFileSequenceCounter = 0;
  char  sz80Filename[84];
  WSADATA wsaData;
  SOCKET  ListeningSocket = INVALID_SOCKET;
  SOCKET  AcceptedSocket  = INVALID_SOCKET;
  SOCKADDR_IN sin_Listen;   // why are theses winsock headers so totally different
  SOCKADDR_IN sin_Remote;   // and incompatible with dozens of TCP examples floating about ?
  FILE *f_out;
  BOOL continue_receiving, continue_listening;
  DWORD dw;

  int nAddrSize;
  int nCode;
  int retval, iWSAErrorCode;
  long nReadBytes;
  char c;

  show_help();

  // Start Winsock up
  if( (nCode = WSAStartup(MAKEWORD(1, 1), &wsaData)) != 0)
   { printf("Sorry, winsock won't cooperate.\r\n");
     getch();
     return 255;
   }

  // Set up a listener on the given interface and port, returning the
  // listening socket if successful; if not, returns INVALID_SOCKET.
  nInterfaceAddr = inet_addr(pcAddress);
  if (nInterfaceAddr != INADDR_NONE)
   {
     ListeningSocket = socket(AF_INET, SOCK_STREAM, 0);
     if (ListeningSocket != INVALID_SOCKET)
      {
        sin_Listen.sin_family      = AF_INET;
        sin_Listen.sin_addr.s_addr = INADDR_ANY;
        sin_Listen.sin_port        = htons( nPort ); // some cretino forgot the 'htons' in his 'tutorial'..
        if (bind(ListeningSocket, (struct sockaddr*)&sin_Listen, sizeof(SOCKADDR_IN)) != SOCKET_ERROR)
         {
           listen(ListeningSocket, 1);
           continue_listening = TRUE;           
           while( continue_listening )
            {
              printf("\r\n Listening on port %d", (int)nPort );

              // Wait for a connection on the given socket.
              // And yes, blocking calls are TERRIBLY STUPID .. but keep it simple here.


              nAddrSize = sizeof(sin_Remote);
              AcceptedSocket = accept(ListeningSocket, (SOCKADDR*)&sin_Remote, &nAddrSize);
              // > If no error occurs, accept returns a value of type SOCKET
              // > which is a descriptor for the accepted socket.
              // > Otherwise, a value of INVALID_SOCKET is returned,
              // > and a specific error code can be retrieved by calling WSAGetLastError.
              if( AcceptedSocket != INVALID_SOCKET )
               { printf("\r\n Accepted connection; making the socked non-blocking." );

                 // Make this socket "non-blocking", so we can detect timeouts here
                 // without the fancy select-stuff or turning this module
                 // into a bloated multi-threading monsteer.
                 // IO-Control FIONBIO: 'Enable or disable nonblocking mode on socket' .
                 dw = 1;  // nonzero for nonblocking mode
                 retval = ioctlsocket( AcceptedSocket, FIONBIO, &dw );
                 if(retval == SOCKET_ERROR)
                  { iWSAErrorCode = WSAGetLastError();
                    closesocket( AcceptedSocket );
                    AcceptedSocket = INVALID_SOCKET;
                    printf("\r\n failed to make the socket non-blocking." );
                  }
               }

             if( AcceptedSocket != INVALID_SOCKET )
               { sprintf( sz80Filename, "test%03d.ogg", iFileSequenceCounter );
                 printf("\r\nCreating output file %s for logging.", sz80Filename );
                 f_out = fopen( sz80Filename, "w+b" ); // overwrite existing file, and use BINARY mode
                 if( f_out == NULL )
                  { printf("\r\n fopen failed." );
                  }
                 else
                  { ++iFileSequenceCounter;
                    // Read data from client, until terminated by user (CTRL-C or 'Q')
                    //  or until nothing happened for more than a few seconds.
                    c = 0;
                    continue_receiving = TRUE;
                    while(c!='q' && c!='Q' && continue_receiving)
                     {
                       nReadBytes = recv(AcceptedSocket, acReadBuffer, sizeof(acReadBuffer), 0);
     // Quoted from some documentation by microsoft  :
     // > If no error occurs, recv returns the number of bytes received.
     // > If the connection has been gracefully closed, the return value
     // > is zero. Otherwise, a value of SOCKET_ERROR is returned,
     // > and a specific error code can be retrieved by calling WSAGetLastError.
     // > If no incoming data is available at the socket,
     // > the recv call waits for data to arrive
     // > unless the socket is nonblocking. In this case, a value of
     // > SOCKET_ERROR is returned with the error code set to WSAEWOULDBLOCK.
     // > The select, WSAAsyncSelect, or WSAEventSelect calls can be used
     // > to determine when more data arrives.
                       if( nReadBytes == 0 ) // "connection has been gracefully closed" ?
                        { printf("\r\n recv() returned ZERO - \"connection gracefully closed\" ?");
                          continue_receiving = FALSE;
                        }
                       else if( nReadBytes > 0 )
                        { // Dump all received bytes into a plain old disk file
                          printf("\r\n recv() returned %d bytes.", (int)nReadBytes );
                          fwrite( acReadBuffer, nReadBytes, 1, f_out );
                        }
                       else if (nReadBytes == SOCKET_ERROR)
                        { // Is this REALLY an error, or just "nothing received" ?
                          // (one would expect return-value = ZERO when 'nothing read' from the input stream,
                          //  but with winsock / recv(), everything is different.
                          //  Or maybe the documentation is just LOUSY .. )
                          iWSAErrorCode = WSAGetLastError();
                          if( iWSAErrorCode == WSAEWOULDBLOCK )  // winsock's way of saying 'nothing received' ?
                           { // No, not really. As usual, the documentation is poor.
                             // recv() often returned ZERO when nothing was received,
                             // instead of returning SOCKET_ERROR + WSAEWOULDBLOCK .
                             // If the call of recv() "would block", wait for a while:
                             Sleep( 100 ); // "sleep" means "give the CPU to someone else for X milliseconds"
                           }
                          else // not just 'nothing received' but a REAL ERROR.  Which ?
                           { printf("\r\n recv() failed, error code #%d .",(int)iWSAErrorCode );
                             continue_receiving = FALSE;
                           }
                        }
                       if( kbhit() )
                        { c = getch();
                          switch( c )
                           { case 'h':
                             case 'H':
                                show_help();
                                break;
                             case 'p':
                                wait_n_seconds_with_indicator( 10 );
                                break;
                             case 'P':
                                wait_n_seconds_with_indicator( 30 );
                                break;
                             case 'q':
                             case 'Q': // quit
                                continue_receiving = FALSE;
                                continue_listening = FALSE;
                                break;
                             case 'r':
                             case 'R': // reset connection, but continue listening
                                continue_receiving = FALSE;
                                continue_listening = TRUE;
                                break;

                           } // end switch (c)
                        } // end if < key pressed >
                     }
                    printf("\r\n Escaped from RX-waiting-loop" );
                    fclose( f_out );
                  } // end if < output-file successfully opened >
               }
              else
               { printf("\r\n accept() failed (%d)",(int)WSAGetLastError() );
               }

              if( AcceptedSocket != INVALID_SOCKET )
               { closesocket( AcceptedSocket );
                 AcceptedSocket = INVALID_SOCKET;
               }
            } // end while( continue_listening )
         }
        else
         { printf("\r\n bind() failed (%d)",(int)WSAGetLastError() );
         }

       } // end if < socket() ok >
      else
       { printf("\r\n socket() failed.");
       }
    } // end if < inet_addr() ok >

  if( ListeningSocket != INVALID_SOCKET )
   { closesocket( ListeningSocket );
     ListeningSocket = INVALID_SOCKET;
   }


  // Shut Winsock back down and take off.
  WSACleanup();
  printf("\r\nPress any key to continue.");
  getch();
  return 0;
}
//---------------------------------------------------------------------------
