/*==========================================================================
  Filename: Cango2v35.js
  Rev: 2
  By: A.R.Collins
  Description: A graphics framework for the canvas element.
  License: Released into the public domain
  latest version at
  <http://www/arc.id.au/CanvasGraphics.html>
  Report bugs to tony at arc.id.au

  Date   |Description                                          |By
  --------------------------------------------------------------------------
  14Oct12 Rev 1.00 First release based on Cango0v43             ARC
  23Oct12 bugfix: xRef,yRef not scaled in most compile methods
          don't pre-scale xRef, yRef in compile methods         ARC
  24Oct12 Give Obj2D methods to simplify compile methods        ARC
  25Oct12 Eliminated group etc. Just use arrays
          Renamed movePoints etc to translate, rotate, scale
          Pass render arguments to color functions              ARC
  26Oct12 Restored the xRef,yRef parameters to compileSvg cmds
          bugfix: restore scaling of font weight                ARC
  27Oct12 Removed iso commands - use 'iso' parameter instead
          bugfix: obj.translate and compileSvg used opposite
          sign conventions - switch compileSvg                  ARC
  11Nov12 removed Object.clone to avoid conflicts with 3rd
          party code (eg ACE editor)                            ARC
  19Nov12 Added to toWorldCoords                                ARC
  20Nov12 Added Obj2D.path to hold 'as rendered' coords
          Added Obj2D.dragNdrop, to hold Drag2D obj             ARC
  24Nov12 Added support for dragNdrop for rotated images
          Removed setRotation clearRotation capability          ARC
  25Nov12 Re-wrote paintPath to better use position transform
          Drag event handlers called in scope of dragNdrop obj
          Re-wrote paintImg to clarify draw origin offsets      ARC
  26Nov12 Added dragNdrop capability to TEXT objects            ARC
  27Nov12 Added Drag2D object, changed .path to .pxOutline
          Added 'as rendered' pxOrg (needed by enableDrag)      ARC
  29Nov11 Added parent as a Drag2D property                     ARC
          Released as Cango2v00                                 ARC
  30Nov12 Added Obj2D.setStrokeWidth
          Made drawPath scl factor permanent, used Obj2D.scale
          Made drawShape scl factor permanent, used Obj2D.scale ARC
  20Dec12 Enabled compileShape to accept numeric data           ARC
  02Jan13 Added stopAnimation,
          Removed animTranslate, animRotate, animScale          ARC
  13Jan13 Changed the shapeDefs to match Cango3D                ARC
  24Jan13 Added cross and ex to shapeDefs                       ARC
  31Jan13 bugfix: multiple Cango contexts over-writing
          mousedown event handler, fixed with global _draggable ARC
  16Feb13 Added Obj2D.appendPath and Obj2D.revWinding           ARC
  19Feb13 Added Obj2D.clipPath                                  ARC
  26Feb13 Use off-screen buffer and bitblt to screen canvas     ARC
  27Feb13 Added shim for isArray, and used it throughout        ARC
  28Feb13 bugfix: bitblt buffer after clearCanvas etc           ARC
  01Mar13 Allow optional direct drawing if node passed in
          Remove useless clearRect before bitblt                ARC
  02Mar13 Added support for multiple animations in one context  ARC
  03Mar13 Added deleteAnimation and uniqueId methods
          reverted to non-buffered drawing                      ARC
  04Mar13 Added pause and step to animations                    ARC
  05Mar13 Added setStepTIme                                     ARC
  06Mar13 Buffer animation drawing and bitblt completed frames  ARC
  10Mar13 Use setProperty and setPropertyDefault methods        ARC
  11Mar13 Added scl parameter to compilePath and compileShape
          Added lineWidth as parameter to compilePath
          dropped strokeColor parm from drawShape
          changed order of parms in drawPath and drawShape
          Added fontWeight and color as compileText parms       ARC
  12Mar13 Added fontSize and fontWeight to settable properties  ARC
  27Mar13 Use native canvasText                                 ARC
  02Apr13 Give clearCanvas an optional fill color               ARC
  28Apr13 bugfix: compile 's' command missing +x, +y            Ales Seifert
  01May13 Make strokeWidth scale with any path scaling          ARC
  31May13 bugfix: typo in clipPath                              ARC
  01Jun13 Use "TEXT","IMG","PATH","SHAPE" instead of objType
          bugfix: _resized etc should be Objects not Arrays
          Many style changes to make JSLint happy               ARC
  02Jun13 Cursor position and csr offsets all in world coords
          bugfix: no check if draggable already in _draggables  ARC
  ==========================================================================*/

// ======================================== GLOBALS ===========================================
  var _resized = {};   // keep track of which canvases are initialised
  var _buf = {};       // each screen canvas will have an off-screen buffer
  var _draggable = {}; // array of Obj2Ds that are draggable for each canvas

// ======================================== SHIMS =============================================
  (function()
  {
    /*-----------------------------------------------------------------------------------------
     * http://paulirish.com/2011/requestanimationframe-for-smart-animating/
     * http://my.opera.com/emoller/blog/2011/12/20/requestanimationframe-for-smart-er-animating
     * requestAnimationFrame polyfill by Erik Möller
     * fixes from Paul Irish and Tino Zijdel
     *----------------------------------------------------------------------------------------*/
    var lastTime = 0;
    var vendors = ['ms', 'moz', 'webkit', 'o'];
    var x;
    for(x = 0; x < vendors.length && !window.requestAnimationFrame; ++x)
    {
      window.requestAnimationFrame = window[vendors[x]+'RequestAnimationFrame'];
      window.cancelAnimationFrame = window[vendors[x]+'CancelAnimationFrame']
                                 || window[vendors[x]+'CancelRequestAnimationFrame'];
    }

    if (!window.requestAnimationFrame)
    {
      window.requestAnimationFrame = function(callback)
      {
        var currTime = new Date().getTime();
        var timeToCall = Math.max(0, 16 - (currTime - lastTime));
        var id = window.setTimeout(function() { callback(currTime + timeToCall); },
          timeToCall);
        lastTime = currTime + timeToCall;
        return id;
      };
    }
    if (!window.cancelAnimationFrame)
    {
      window.cancelAnimationFrame = function(id)
      {
        clearTimeout(id);
      };
    }

    if (!Date.now)
    {
      Date.now = function now()
      {
        return new Date().getTime();
      };
    }

    /* create a copy (not just a reference) of an object */
    if (!Object.prototype.clone)
    {
      Object.prototype.clone = function()
      {
        var newObj = (this instanceof Array) ? [] : {};
        var i;
        for (i in this)
        {
          if (this.hasOwnProperty(i))
          {
            if (i == 'clone')
            {
               continue;
            }
            if (this[i] && typeof this[i] === "object")
            {
              newObj[i] = this[i].clone();
            }
            else
            {
              newObj[i] = this[i];
            }
          }
        }
        return newObj;
      };
    }

    if (!window.isArray)
    {
      isArray = function(obj)
      {
        return Object.prototype.toString.call(obj) === '[object Array]';
      };
    }

    if (!window.isNumber)
    {
      isNumber = function(o)
      {
        return !isNaN(o) && o !== null && o !== "" && o !== false;
      };
    }

    if (!window.addLoadEvent)
    {
      // simple add event handler that has the handlers called in the sequence that they were set
      addLoadEvent = function(obj, func)
      {
      	var oldonload = obj.onload;

      	if (typeof(obj.onload) != "function")
        {
          obj.onload = func;
        }
      	else
        {
        	obj.onload = function(){ oldonload(); func(); };
        }
      };
    }

    if (!window.addEvent)
    {
      addEvent = function(element, eventType, handler)
      {
        if (element.attachEvent)
        {
         return element.attachEvent('on'+eventType, handler);
        }
        return element.addEventListener(eventType, handler, false);
      };
    }

    if (!window.removeEvent)
    {
      removeEvent = function(element, eventType, handler)
      {
       if (element.removeEventListener)
       {
          element.removeEventListener (eventType, handler, false);
       }
       if (element.detachEvent)
       {
          element.detachEvent ('on'+eventType, handler);
       }
      };
    }

    if (!Array.prototype.contains)
    {
      Array.prototype.contains = function(obj)
      {
        var i = this.length;
        while (i--)
        {
          if (this[i] === obj)
          {
            return true;
          }
        }
        return false;
      };
    }

    if (!Array.prototype.map)
    {
      Array.prototype.map = function(fun, thisArg)
      {
        var i;
        var len = this.length;
        if (typeof fun != "function")
        {
          throw new TypeError();
        }
        var res = [];
        var thisp = thisArg;
        for (i = 0; i < len; i++)
        {
          if (this.hasOwnProperty(i))
          {
            res[i] = fun.call(thisp, this[i], i, this);
          }
        }
        return res;
      };
    }

  }());

  function DrawCmd(cmdStr, coords)   // canvas syntax draw commands
  {
    this.drawFn = cmdStr;       // String version of the canvas command to call
    this.parms = coords || [];  // array of parameters to pass to drawFn
  }

  function Obj2D(cgo, commands, objtype, isotropic, fillColor, strokeColor, lorg, dragObj)
  {
    this.drawCmds = commands;      // array of DrawCmd objects holding cmds and world coords (or an Img object)
    this.pxOutline = [];           // array of DrawCmd objects holding 'as rendered' cmds and pixel coords
    this.pxOrgX = 0;               // pixels coords of the 'as rendered' drawing origin
    this.pxOrgY = 0;               //     "
    this.type = "SHAPE";           // enum of type to instruct the render method
    this.iso = false;              // true = maintain aspect ratio
    this.strokeCol = "black";      // renderer will stroke a path in this color
    this.fillCol = "gray";         // only used if type = SHAPE
    this.strokeWidth = 1;          // in case fat outline is wanted for Path or Shape outline
    this.strokeCap = "butt";       // freeze current style in case something fancy is wanted
    this.width = 0;                // only used for type = IMG, TEXT, set to 0 until image loaded
    this.height = 0;               //     "
    this.imgX = 0;                 // TEXT & IMG use these for obj.translate, obj.rotate, obj.scale
    this.imgY = 0;                 //     "
    this.imgXscale = 1;            //     "
    this.imgYscale = 1;            //     "
    this.imgDegs = 0;              //     "
    this.fontSize = 10;            // fontSize in points (TEXT only)
    this.fontWeight = 400;         // fontWeight 100..900 (TEXT only)
    this.lorg = lorg || 1;         // only used for type = IMG and TEXT
    this.dragNdrop = dragObj || null;

    if (cgo)
    {
      this.type = objtype || "SHAPE";
      this.iso = (isotropic == true);
      // check for iso error
      if ((this.type == "IMG")||(this.type == "TEXT"))
      {
        this.iso = true;
      }

      if ((fillColor !== undefined)&&(fillColor != null))
      {
        this.fillCol = fillColor;
      }
      else
      {
        this.fillCol = cgo.paintCol;
      }

      if ((strokeColor !== undefined)&&(strokeColor != null))
      {
        this.strokeCol = strokeColor;
      }
      else if (this.type == "SHAPE")
      {
        this.strokeCol = this.fillCol; // shapes default to stroke and fill the same
      }
      else
      {
        this.strokeCol = cgo.penCol;   // path and text default to current pen color
      }

      this.strokeWidth = cgo.penWid;
      this.strokeCap = cgo.lineCap;
      if (this.type == "TEXT")
      {
        this.strokeCap = 'round';                // text must be stroked round to look good
      }
    }
    if (this.dragNdrop != null)
    {
      this.dragNdrop.parent = this;    // give dragNdrop callBacks easy access to the object
    }
  }

  Obj2D.prototype.translate = function(x, y)
  {
    var cmd;
    var i, j;
    if ((this.type == "IMG")||(this.type == "TEXT"))
    {
      // no points to shift just remeber the offset to use when rendering
      this.imgX += x;     // IMG and TEXT types do a translate during render
      this.imgY += y;
    }
    else
    {
      for (i=0; i<this.drawCmds.length; i++)
      {
        cmd = this.drawCmds[i];  // for clarity
        for (j=0; j<cmd.parms.length/2; j++)
        {
          cmd.parms[2*j] += x;
          cmd.parms[2*j+1] += y;
        }
      }
    }
  };

  Obj2D.prototype.rotate = function(degs)
  {
    var cmd;
    var A = Math.PI*degs/180.0;   // radians
    var sinA = Math.sin(A);
    var cosA = Math.cos(A);
    var x, y, i, j;

    if ((this.type == "IMG")||(this.type == "TEXT"))
    {
      // no points to shift just remember the value to use when rendering
      this.imgDegs = degs;
    }
    else
    {
      for (i=0; i<this.drawCmds.length; i++)
      {
        cmd = this.drawCmds[i];  // for clarity
        for (j=0; j<cmd.parms.length/2; j++)
        {
          x = cmd.parms[2*j];
          y = cmd.parms[2*j+1];
          cmd.parms[2*j] = x*cosA - y*sinA;
          cmd.parms[2*j+1] = x*sinA + y*cosA;
        }
      }
    }
  };

  Obj2D.prototype.scale = function(xScl, yScl)
  {
    var cmd;
    var xScale = xScl || 1;
    var yScale = yScl || xScale;   // default to isotropic scaling
    var i, j;

    if ((this.type == "IMG")||(this.type == "TEXT"))
    {
      // no points to shift just remeber values to use when rendering
      this.imgXscale *= xScale;
      this.imgYscale *= yScale;
      this.imgX *= xScale;
      this.imgY *= yScale;
    }
    else
    {
      for (i=0; i<this.drawCmds.length; i++)
      {
        cmd = this.drawCmds[i];  // for clarity
        for (j=0; j<cmd.parms.length/2; j++)
        {
          cmd.parms[2*j] *= xScale;
          cmd.parms[2*j+1] *= yScale;
        }
      }
      if (this.type == "PATH")                // let line thickness of Paths scale too
      {
        this.strokeWidth *=xScale;
      }
    }
  };

  Obj2D.prototype.appendPath = function(obj, delMove)
  {
    if ((this.type == "IMG")||(this.type == "TEXT")||(obj.type == "IMG")||(obj.type == "TEXT"))
    {
      return;
    }
    if (delMove)  // delete the inital 'moveTo' command
    {
      this.drawCmds = this.drawCmds.concat(obj.drawCmds.slice(1));
    }
    else
    {
      this.drawCmds = this.drawCmds.concat(obj.drawCmds);
    }
  };

  Obj2D.prototype.revWinding = function()
  {
    function revPairs(ary)
    {
      if (ary.length == 2)
      {
        return ary;
      }
      var revAry = [];
      var i, j;
      for (i=0, j=ary.length/2-1; j>=0; i++, j--)
      {
        revAry[2*i] = ary[2*j];
        revAry[2*i+1] = ary[2*j+1];
      }
      return revAry;
    }
    if ((this.type == "IMG")||(this.type == "TEXT"))
    {
      return;
    }
    // reverse the direction of drawing around a path, stops holes in shapes being filled
    var cmds;
    var zCmd = null;
    if (this.drawCmds[this.drawCmds.length-1].drawFn.toUpperCase() == "Z")
    {
      cmds = this.drawCmds.slice(0, -1);  // leave off Z
      zCmd = this.drawCmds.slice(-1);
    }
    else
    {
      cmds = this.drawCmds.slice(0);  // copy the whole array
    }
    var revCmds = [];
    // now step back along the path
    var j = cmds.length-1;
    var len = cmds[j].parms.length;
    var dParms = [cmds[j].parms[len-2], cmds[j].parms[len-1]];
    var dCmd = new DrawCmd("moveTo", dParms);
    revCmds.push(dCmd);
    cmds[j].parms = cmds[j].parms.slice(0,-2);  // weve used the last point so slice it off
    while (j>0)
    {
      dParms = revPairs(cmds[j].parms);
      len = cmds[j-1].parms.length;
      dParms.push(cmds[j-1].parms[len-2], cmds[j-1].parms[len-1]); // push last of next cmd
      dCmd = new DrawCmd(cmds[j].drawFn, dParms);
      revCmds.push(dCmd);
      cmds[j-1].parms = cmds[j-1].parms.slice(0,-2);  // weve used the last point so slice it off
      j--;
    }
    // add the 'z' if it was a closed path
    if (zCmd)
    {
      revCmds.push(zCmd);
    }

    this.drawCmds = revCmds;
  };

  Obj2D.prototype.enableDrag = function(drag)
  {
    this.dragNdrop = drag;
    // fill in the Drag2D properties for use by callBacks
    this.dragNdrop.parent = this;
    // the Obj2D may already have been rendered, save the 'as rendered' drawing origin
    var orgWC = this.toWorldCoords(this.pxOrgX, this.pxOrgY);
    this.dragNdrop.xOrg = orgWC.x;
    this.dragNdrop.yOrg = orgWC.y;
    // include this in objects to be checked on mousedown
    // the Drag2D has the cango context saved as 'this.cgo'
    _draggable[drag.cgo.cId].push(this);
  };

  Obj2D.prototype.disableDrag = function()
  {
    function getIndex(ary, obj)
    {
      var i, j;
      for (i=0, j=ary.length; i<j; i++)
      {
        if (ary[i] === obj)
        {
          return i;
        }
      }
      return -1;
    }
    if (!this.dragNdrop)
    {
      return;
    }
    // remove this object from array to be checked on mousedown
    // the Drag2D has the cango context saved as 'this.cgo'
    var aidx = getIndex(_draggable[this.dragNdrop.cgo.cId], this);
    _draggable[this.dragNdrop.cgo.cId].splice(aidx, 1);
    this.dragNdrop = null;
  };

  Obj2D.prototype.setProperty = function(propertyName, value)
  {
    if ((typeof propertyName !== "string")||(value === undefined)||(value == null))
    {
      return;
    }

    switch (propertyName.toLowerCase())
    {
      case "fillcolor":
        if ((typeof value != "string")&&(typeof value != "object"))  // gradient will be an object
        {
          return;
        }
        this.fillCol = value;
        break;
      case "strokecolor":
        if ((typeof value != "string")&&(typeof value != "object"))  // gradient will be an object
        {
          return;
        }
        this.strokeCol = value;
        break;
      case "strokewidth":
        this.strokeWidth = value;
        break;
      case "linecap":
        if (typeof value != "string")
        {
          return;
        }
        if ((value == "butt")||(value =="round")||(value == "square"))
        {
          this.strokeCap = value;
        }
        break;
      case "fontsize":
        this.fontSize = Math.abs(value);
        break;
      case "fontweight":
        if ((typeof value == "string")||((value>=100)&&(value<=900)))
        {
          this.fontWeight = value;
        }
        break;
      default:
        return;
    }
  };

  Obj2D.prototype.dup = function()
  {
    function cloneDrawCmds(orgCmds)
    {
      var newcmds = (isArray(orgCmds)) ? [] : {};
      var i;
      for (i in orgCmds)
      {
        if (orgCmds[i] && typeof orgCmds[i] == "object")
        {
          newcmds[i] = cloneDrawCmds(orgCmds[i]);
        }
        else
        {
          newcmds[i] = orgCmds[i];
        }
      }
      return newcmds;
    }

    var newObj = new Obj2D();

    newObj.drawCmds = cloneDrawCmds(this.drawCmds);   // might be IMG or Array
    newObj.type = this.type;
    newObj.iso = this.iso;
    newObj.strokeCol = this.strokeCol;
    newObj.fillCol = this.fillCol;
    newObj.strokeWidth = this.strokeWidth;
    newObj.strokeCap = this.strokeCap;
    newObj.width = this.width;
    newObj.height = this.height;
    newObj.imgX = this.imgX;
    newObj.imgY = this.imgY;
    newObj.imgXscale = this.imgXscale;
    newObj.imgYscale = this.imgYscale;
    newObj.imgDegs = this.imgDegs;
    newObj.lorg = this.lorg;

    return newObj;
  };

//===============================================================================

  var shapeDefs = {'circle':['M', -0.5, 0, 'a', 0.5, 0.5, 0, 1, 0, 1, 0, 0.5, 0.5, 0, 1, 0, -1, 0],
                   'square':['M', 0.5, -0.5, 'l', 0, 1, -1, 0, 0, -1, 'z'],
                   'triangle':['M', 0.5, -0.289, 'l', -0.5, 0.866, -0.5, -0.866, 'z'],
                   'cross':['M', -0.5, 0, 'l', 1, 0, 'M', 0, -0.5, 'l', 0, 1],
                   'ex':['M', -0.3535,-0.3535, 'L',0.3535,0.3535, 'M',-0.3535,0.3535, 'L',0.3535,-0.3535]};

  function Cango(canvasId)
  {
    this.cId = canvasId;
    this.cnvs = document.getElementById(canvasId);
    if (this.cnvs == null)
    {
      alert("can't find canvas "+canvasId);
      return;
    }
    this.rawWidth = this.cnvs.offsetWidth;
    this.rawHeight = this.cnvs.offsetHeight;
    this.aRatio = this.rawWidth/this.rawHeight;
    this.buffered = true;// draw animation to off-screen buffer then bitblt complete frames
    if (!_resized.hasOwnProperty(this.cId))
    {
      // make canvas native aspect ratio equal style box aspect ratio.
      // Note: rawWidth and rawHeight are floats, assignment to ints will truncate
      this.cnvs.setAttribute('width', this.rawWidth);    // this actually resets number of graphics pixels
      this.cnvs.setAttribute('height', this.rawHeight);  // don't use style for this
      // create element for this canvas, this prevents resize for each Cango instance
      _resized[this.cId] = true;
      // create off screen drawing buffer
      if (this.buffered)
      {
        _buf[this.cId] = document.createElement('canvas');      // create buffer in memory
        _buf[this.cId].setAttribute('width', this.rawWidth);    // set number of graphics pixels
        _buf[this.cId].setAttribute('height', this.rawHeight);  // to match screen canvas
      }
      // create an array to hold all the draggable objects for each canvas
      _draggable[this.cId] = [];
    }
    if (this.buffered)
    {
      this.bufCtx = _buf[this.cId].getContext('2d');  // animation drawing done off screen
    }
    this.ctx = this.cnvs.getContext('2d');    // draw direct to screen canvas
    this.vpW = this.rawWidth;         // vp width in pixels (default to full canvas size)
    this.vpH = this.rawHeight;        // vp height in pixels
    this.vpLLx = 0;                   // vp lower left from canvas left in pixels
    this.vpLLy = this.rawHeight;      // vp lower left from canvas top
    this.xscl = this.rawWidth/100;    // world x axis scale factor, default: canvas width = 100 (equ. vp units)
    this.yscl = -this.rawWidth/100;   // world y axis scale factor, +ve up, canvas height =100*aspect ratio
    this.xoffset = 0;                 // world x origin offset from viewport left in pixels
    this.yoffset = 0;                 // world y origin offset from viewport bottom in pixels
    // *** to move to world coord x,y ***
    // 1. from pixel x origin (canvas left) add vpLLx (gets to viewport left)
    // 2. add xoffset to get to pixel location of world X origin
    // 3. add x*xscl pixels to get to world x location.
    // ==> x (in world coords) == vpLLx + xoffset + x*xscl (pixels location on canvas)
    // ==> y (in world coords) == vpLLy + yoffset + y*xscl (pixels location on canvas).
    this.ctx.textAlign = "left";     // all offsets re handled in code so dragNdrop works
    this.ctx.textBaseline = "top";

    this.penCol = "rgba(0, 0, 0, 1.0)";        // black
    this.penWid = 1;                  // pixels
    this.lineCap = "butt";
    this.paintCol = "rgba(128, 128, 128, 1.0)";  // gray
    this.fontSize = 10;               // pt
    this.fontWeight = 400;            // 100..900, 400 = normal,700 = bold
    this.fontFamily = "'Lucinda Console', monospace";
    this.animations = [];             // array of animations to be drawn
    this.timer = null;                // need to save the rAF id for cancelling
    this.modes = {PAUSED:1, STOPPED:2, PLAYING:3, STEPPING:4};     // animation modes
    this.animMode = this.modes.STOPPED;
    this.prevAnimMode = this.modes.STOPPED;
    this.startTime = 0;   // animation start time (relative to 1970)
    this.currTime = 0;    // timestamp of frame on screen
    this.stepTime = 50;   // animation step time interval (in msec)
    this.unique = 0;                  // counter to generate unique IDs

    var savThis = this;

    this.cnvs.onmousedown = function(evt)
    {
      var event = evt || window.event;
      var crsPos = savThis.getCursorPos(event);  // savThis is any Cango ctx on the canvas
      function hitTest(pathObj)
      {
        var ctx = pathObj.dragNdrop.cgo.ctx;  // cgo is saved in each Drag2D
        // create the path (don't stroke it - no-one will see) to test for hit
        var i;
        ctx.beginPath();
        for (i=0; i<pathObj.pxOutline.length; i++)
        {
          ctx[pathObj.pxOutline[i].drawFn].apply(ctx, pathObj.pxOutline[i].parms);
        }
        // for diagnostics on hit region, un-comment the ctx.stroke(); below
        // ctx.stroke();
        return ctx.isPointInPath(crsPos.x, crsPos.y);
      }
      // run through all the registered objects and test if cursor pos is in their path
      var j;
      for (j=0; j<_draggable[savThis.cId].length; j++)
      {
        if (hitTest(_draggable[savThis.cId][j]))
        {
          // call the grab handler for this object
          _draggable[savThis.cId][j].dragNdrop.grab(event);
        }
      }
    };
  }

  Cango.prototype.uniqueId = function()
  {
    this.unique++;
    return this.cId+"_"+this.unique;    // return a unique string
  };

  Cango.prototype.toViewportCoords = function(x, y)
  {
    // transform x,y in world coords to viewport coords (x axis 0 to 100, y axis same scale factor)
    var xPx = this.vpLLx+this.xoffset+x*this.xscl;
    var yPx = this.vpLLy+this.yoffset+y*this.yscl;

    return {x: 100*xPx/this.rawWidth, y: 100*(this.rawHeight-yPx)/this.rawWidth};
  };

  Cango.prototype.toPixelCoords = function(x, y)
  {
    // transform x,y in world coords to canvas pixel coords (top left is 0,0 y axis +ve down)
    var xPx = this.vpLLx+this.xoffset+x*this.xscl;
    var yPx = this.vpLLy+this.yoffset+y*this.yscl;

    return {x: xPx, y: yPx};
  };

  Cango.prototype.toWorldCoords = function(xPx, yPx)
  {
    // transform xPx,yPx in raw canvas pixels to world coords (lower left is 0,0 +ve up)
    var xW = (xPx - this.vpLLx - this.xoffset)/this.xscl;
    var yW = (yPx - this.vpLLy - this.yoffset)/this.yscl;

    return {x: xW, y: yW};
  };

  Cango.prototype.getCursorPos = function(evt)
  {
    // pass in any mouse event, returns the position of the cursor in raw pixel coords
    var event = evt||window.event;
    var rect = this.cnvs.getBoundingClientRect();

    return {x: event.clientX - rect.left, y: event.clientY - rect.top};
  };

  Cango.prototype.clearCanvas = function(fillColor)
  {
    /* clearCanvas() sets the background color and fills the canvas to that color */
    if (fillColor != undefined)
    {
      this.ctx.fillStyle = fillColor;
      this.ctx.fillRect(0, 0, this.rawWidth, this.rawHeight);
    }
    else
    {
      this.ctx.clearRect(0, 0, this.rawWidth, this.rawHeight);
    }
    // all drawing erased, but graphics contexts remain intact
    // clear the draggable array, draggables put back when rendered
    _draggable[this.cId] = [];
  };

  Cango.prototype.setViewport = function(lowerLeftX, lowerLeftY, w, h)
  {
    if (h != undefined)
    {
      this.vpW = 0.01*w*this.rawWidth;
      this.vpH = 0.01*h*this.rawWidth;
      this.vpLLx = 0.01*lowerLeftX*this.rawWidth;
      this.vpLLy = this.rawHeight-0.01*lowerLeftY*this.rawWidth;
    }
    else
    {
      this.vpW = this.rawWidth;
      this.vpH = this.rawHeight;
      this.vpLLx = 0;
      this.vpLLy = this.rawHeight;
    }
    this.setWorldCoords();     // if new viewport, world coords are garbage, so reset to defaults
  };

  Cango.prototype.fillViewport = function(fillColor)
  {
    var newCol = this.paintCol;
    // set background color and fill the viewport to that color or gradient
    if ((fillColor != undefined)&&(fillColor != null))
    {
      newCol = fillColor;
    }
    this.ctx.save();
    this.ctx.fillStyle = newCol;
    this.ctx.fillRect(this.vpLLx, (this.vpLLy-this.vpH), this.vpW, this.vpH); // fill: top, left, wid, hgt
    this.ctx.restore();
  };

  Cango.prototype.setWorldCoords = function(leftX, rightX, lowerY, upperY)
  {
    if (upperY != undefined)
    {
      this.xscl = this.vpW/(rightX-leftX);
      this.yscl = -(this.vpH/(upperY-lowerY));
      this.xoffset = -leftX*this.xscl;
      this.yoffset = -lowerY*this.yscl;
    }
    else
    {
      this.xscl = this.rawWidth/100;    // makes x axis = 100 native units
      this.yscl = -this.rawWidth/100;   // makes y axis = 100*aspect ratio ie. square pixels
      this.xoffset = 0;
      this.yoffset = 0;
    }
  };

  Cango.prototype.setPropertyDefault = function(propertyName, value)
  {
    if ((typeof propertyName != "string")||(value == undefined)||(value == null))
    {
      return;
    }

    switch (propertyName.toLowerCase())
    {
      case "fillcolor":
        if ((typeof value != "string")&&(typeof value != "object"))  // gradient will be an object
        {
          return;
        }
        this.paintCol = value;
        break;
      case "strokecolor":
        if ((typeof value != "string")&&(typeof value != "object"))  // gradient will be an object
        {
          return;
        }
        this.penCol = value;
        break;
      case "strokewidth":
        this.penWid = value;
        break;
      case "linecap":
        if (typeof value != "string")
        {
          return;
        }
        if ((value == "butt")||(value =="round")||(value == "square"))
        {
          this.lineCap = value;
        }
        break;
      case "fontfamily":
        if (typeof value != "string")
        {
          return;
        }
        this.fontFamily = value;
        break;
      case "fontsize":
        this.fontSize = value;
        break;
      case "fontweight":
        if ((typeof value == "string")||((value >= 100)&&(value <= 900)))
        {
          this.fontWeight = value;
        }
        break;
      case "steptime":
        if ((value >= 15)&&(value <= 500))
        {
          this.stepTime = value;
        }
        break;
      default:
        return;
    }
  };

  Cango.prototype.linearGradientFill = function(x1, y1, x2, y2, x, y, scl, isotropic)
  {
    var xOfs = x || 0;
    var yOfs = y || 0;
    var xScl = scl || 1;
    var yScl = scl || 1;
    if ((isotropic != undefined)&&(isotropic == 'iso'))
    {
      yScl *= -this.xscl/this.yscl;
    }
    // pixel version of world coordinate parms
    var p1x = xOfs+x1*xScl;
    var p1y = yOfs+y1*yScl;
    var p2x = xOfs+x2*xScl;
    var p2y = yOfs+y2*yScl;

    return this.ctx.createLinearGradient(this.xscl*p1x, this.yscl*p1y, this.xscl*p2x, this.yscl*p2y);
  };

  Cango.prototype.radialGradientFill = function(x1, y1, r1, x2, y2, r2, x, y, scl, isotropic)
  {
    var xOfs = x || 0;
    var yOfs = y || 0;
    var xScl = scl || 1;
    var yScl = scl || 1;
    if ((isotropic != undefined)&&(isotropic == 'iso'))
    {
      yScl *= -this.xscl/this.yscl;
    }
    // world coordinate parms (equivalent to compile methods)
    var p1x = xOfs+x1*xScl;
    var p1y = yOfs+y1*yScl;
    var p1r = r1*this.xscl*xScl;
    var p2x = xOfs+x2*xScl;
    var p2y = yOfs+y2*yScl;
    var p2r = r2*this.xscl*xScl;

    return this.ctx.createRadialGradient(this.xscl*p1x, this.yscl*p1y, p1r, this.xscl*p2x, this.yscl*p2y, p2r);
  };

  Cango.prototype.render = function(pathObj, x, y, scl, degs)
  {
    var savThis = this;

    function loadCallback(obj, x, y, scl, degs)
    {
      return  function(){
                          savThis._paintImg(obj, x, y, scl, scl, degs);
                        };
    }
    // test for an array or just a single path
    var i;
    if (isArray(pathObj))          // array of Obj2D
    {
      for (i=0; i<pathObj.length; i++)
      {
        if (pathObj[i].type == "IMG")
        {
          if (pathObj[i].width>0)  // image loaded and width set?
          {
            this._paintImg(pathObj[i], x, y, scl, scl, degs);
          }
          else
          {
            addLoadEvent(pathObj[i].drawCmds, loadCallback(pathObj[i], x, y, scl, degs));
          }
        }
        else if (pathObj[i].type == "TEXT")
        {
          this._paintText(pathObj[i], x, y, scl, scl, degs);
        }
        else    // PATH, SHAPE
        {
          this._paintPath(pathObj[i], x, y, scl, scl, degs);
        }
      }
    }
    else
    {
      if (pathObj.type == "IMG")
      {
        if (pathObj.width>0)  // image loaded, and width set?
        {
          this._paintImg(pathObj, x, y, scl, scl, degs);
        }
        else
        {
          addLoadEvent(pathObj.drawCmds, loadCallback(pathObj, x, y, scl, degs));
        }
      }
      else if (pathObj.type == "TEXT")
      {
        this._paintText(pathObj, x, y, scl, scl, degs);
      }
      else    // PATH, SHAPE
      {
        this._paintPath(pathObj, x, y, scl, scl, degs);
      }
    }
  };

  Cango.prototype._paintImg = function(pathObj, x, y, xScl, yScl, degrees)
  {
    // should only be called after image has been loaded into drawCmds
    var img = pathObj.drawCmds;            // this is the place the image is stored in object
    var w = pathObj.width;                 // this is the original requested width in world coords
    var h = pathObj.height;
    var xPos = x || 0;
    var yPos = y || 0;
    var xScale = xScl || 1;
    var yScale = yScl || 1;
    var scale = xScale*pathObj.imgXscale;  // imgXscale is from Obj2D.scale (permanent)
    var dgs = degrees || 0;
    var degs = dgs + pathObj.imgDegs;

    this.ctx.save();   // save the clean ctx
    // move the whole coordinate system to the xPos,yPos
    this.ctx.translate(this.vpLLx+this.xoffset+xPos*this.xscl, this.vpLLy+this.yoffset+yPos*this.yscl);
    if (degs)
    {
      this.ctx.rotate(-degs*Math.PI/180.0);   // rotate
    }
    // now insert the image at the rotation radius from the origin and scaled in width
    this.ctx.drawImage(img, this.xscl*scale*pathObj.imgX, this.xscl*scale*pathObj.imgY,
                            this.xscl*scale*pathObj.width, this.xscl*scale*pathObj.height);
    this.ctx.restore();    // undo the transforms

    // make a hitRegion boundary path around the image to be checked on mousedown
    var ulx = this.xscl*scale*pathObj.imgX;
    var uly = this.xscl*scale*pathObj.imgY;
    var urx = this.xscl*scale*(pathObj.imgX+pathObj.width);
    var ury = uly;
    var lrx = urx;
    var lry = this.xscl*scale*(pathObj.imgY+pathObj.height);
    var llx = ulx;
    var lly = lry;

    var vul = [ulx, uly];
    var vur = [urx, ury];
    var vlr = [lrx, lry];
    var vll = [llx, lly];
    // if the image has been rotated then rotate the bounding box
    if (degs)
    {
      var A = -Math.PI*degs/180.0;   // radians
      var sinA = Math.sin(A);
      var cosA = Math.cos(A);
      var rotXY = function(x,y)
                  {
                    return [x*cosA - y*sinA, x*sinA + y*cosA];
                  };
      vul = rotXY(ulx, uly);
      vur = rotXY(urx, ury);
      vlr = rotXY(lrx, lry);
      vll = rotXY(llx, lly);
    }
    // now form the DrawCmds and translate the vertices
    var drawOrgX = this.vpLLx+this.xoffset+this.xscl*xPos;
    var drawOrgY = this.vpLLy+this.yoffset+this.yscl*yPos;
    var moveXY = function(v)
                  {
                    return [drawOrgX+v[0], drawOrgY+v[1]];
                  };
    pathObj.pxOutline = [];   // start with new array
    pathObj.pxOutline[0] = new DrawCmd('moveTo', moveXY(vul));
    pathObj.pxOutline[1] = new DrawCmd('lineTo', moveXY(vur));
    pathObj.pxOutline[2] = new DrawCmd('lineTo', moveXY(vlr));
    pathObj.pxOutline[3] = new DrawCmd('lineTo', moveXY(vll));
    pathObj.pxOutline[4] = new DrawCmd('closePath');
    pathObj.pxOrgX = this.vpLLx+this.xoffset+xPos*this.xscl;
    pathObj.pxOrgY = this.vpLLy+this.yoffset+yPos*this.yscl;

    if (pathObj.dragNdrop != null)
    {
      // save pixel coords of the drawing origin (often useful in drag n drop)
      pathObj.dragNdrop.xOrg = xPos;
      pathObj.dragNdrop.yOrg = yPos;
      // now push it into Cango.draggable array, its checked by canvas mousedown event handler
      if (!_draggable[this.cId].contains(pathObj))
      {
        _draggable[this.cId].push(pathObj);
      }
    }
  };

  Cango.prototype._paintPath = function(pathObj, x, y, xScl, yScl, degrees)
  {
    // used for type: PATH, SHAPE
    var xPos = x || 0;
    var yPos = y || 0;
    var xScale = xScl || 1;
    var yScale = yScl || 1;
    var degs = degrees || 0;
    var fill = null;
    var pxlCoords = [];
    var i, j;

    this.ctx.save();   // save current context
    this.ctx.translate(this.vpLLx+this.xoffset+xPos*this.xscl, this.vpLLy+this.yoffset+yPos*this.yscl);
    this.ctx.scale(xScale, yScale);
    this.ctx.save();   // save current rotation if any
    if (degs)
    {
      this.ctx.rotate(-degs*Math.PI/180.0);
    }
    this.ctx.beginPath();
    for (i=0; i<pathObj.drawCmds.length; i++)
    {
      for (j=0; j<pathObj.drawCmds[i].parms.length; j+=2)  // convert x,y coord pairs to pixel coords
      {
        pxlCoords[j] = this.xscl*pathObj.drawCmds[i].parms[j];
        if (pathObj.iso)
        {
          pxlCoords[j+1] = -this.xscl*pathObj.drawCmds[i].parms[j+1];
        }
        else
        {
          pxlCoords[j+1] = this.yscl*pathObj.drawCmds[i].parms[j+1];
        }
      }
      this.ctx[pathObj.drawCmds[i].drawFn].apply(this.ctx, pxlCoords); // draw the path
    }
    // undo the rotation before fill so pattern not rotated
    this.ctx.restore();
    // if a SHAPE, fill with color
    if (pathObj.type == "SHAPE")
    {
      // pathObj.fillCol may be a function that generates dynamic color (so call it)
      if (pathObj.fillCol instanceof Function)
      {
        fill = pathObj.fillCol(arguments);
      }
      else
      {
        fill = pathObj.fillCol;
      }
      this.ctx.fillStyle = fill;
      this.ctx.fill();
    }
    // try to avoid calling color function twice (if just filled 'fill' will be a color)
    if ((pathObj.fillCol == pathObj.strokeCol)&&(fill != null))
    {
      this.ctx.strokeStyle = fill;
    }
    else
    {
      // pathObj.fillCol may be a function that generates dynamic color (so call it)
      if (pathObj.strokeCol instanceof Function)
      {
        this.ctx.strokeStyle = pathObj.strokeCol(arguments);
      }
      else
      {
        this.ctx.strokeStyle = pathObj.strokeCol;
      }
    }
    this.ctx.lineWidth = Math.abs(pathObj.strokeWidth); // strokeWidth set at compile time
    this.ctx.lineCap = pathObj.strokeCap;
    this.ctx.stroke();
     // undo the translation and scaling
    this.ctx.restore();

    // construct the 'as rendered' outline (in pixels) for mouse event detection
    // if the path has been rotated set up some tools to rotate the pixel coords
    var A, sinA, cosA, rotXY;
    if (degs)
    {
      A = -Math.PI*degs/180.0;   // radians
      sinA = Math.sin(A);
      cosA = Math.cos(A);
      rotXY = function(x,y)
              {
                return [x*cosA - y*sinA, x*sinA + y*cosA];
              };
    }
    pathObj.pxOutline = [];   // start with new array
    // scale and translate to pixel coords (save these for drag and drop testing)
    pathObj.pxOrgX = this.vpLLx+this.xoffset+xPos*this.xscl;
    pathObj.pxOrgY = this.vpLLy+this.yoffset+yPos*this.yscl;

    var pxCoords, v;
    for (i=0; i<pathObj.drawCmds.length; i++)
    {
      pxCoords = [];
      for (j=0; j<pathObj.drawCmds[i].parms.length; j+=2)      // convert x,y coord pairs to world coords
      {
        pxCoords[j] = xScale*pathObj.drawCmds[i].parms[j]*this.xscl;
        if (pathObj.iso)
        {
          pxCoords[j+1] = yScale*pathObj.drawCmds[i].parms[j+1]*-this.xscl;
        }
        else
        {
          pxCoords[j+1] = yScale*pathObj.drawCmds[i].parms[j+1]*this.yscl;
        }
        if (degs)
        {
          v = rotXY(pxCoords[j], pxCoords[j+1]);
          pxCoords[j] = v[0];
          pxCoords[j+1] = v[1];
        }
        pxCoords[j] += pathObj.pxOrgX;   // position path by adding the coordinates of the drawing origin
        pxCoords[j+1] += pathObj.pxOrgY;
      }
      pathObj.pxOutline.push(new DrawCmd(pathObj.drawCmds[i].drawFn, pxCoords));
    }

    if (pathObj.dragNdrop != null)
    {
      // save pixel coords of the drawing origin (often useful in drag n drop)
      pathObj.dragNdrop.xOrg = xPos;
      pathObj.dragNdrop.yOrg = yPos;
      // now push it into Cango.draggable array, its checked by canvas mousedown event handler
      if (!_draggable[this.cId].contains(pathObj))
      {
        _draggable[this.cId].push(pathObj);
      }
    }
  };

  Cango.prototype._paintText = function(pathObj, x, y, xScl, yScl, degrees)
  {
    // used for type: PATH, SHAPE or TEXT
    var xPos = x || 0;
    var yPos = y || 0;
    var xScale = xScl || 1;
    var yScale = yScl || 1;
    var degs = degrees || 0;
    var fill = null;
    var pxlCoords = [];
    var i, j;
    // translate, rotate, scale must be handled at render time
    degs += pathObj.imgDegs;

    this.ctx.save();   // save current context
    this.ctx.translate(this.vpLLx+this.xoffset+xPos*this.xscl, this.vpLLy+this.yoffset+yPos*this.yscl);
    this.ctx.scale(xScale, yScale);
    this.ctx.save();   // save current rotation if any
    if (degs)
    {
      this.ctx.rotate(-degs*Math.PI/180.0);
    }
    this.ctx.font = pathObj.fontWeight.toString()+" "+pathObj.fontSize+"px "+this.fontFamily;
    // set the fillStyle to strokeColor for text
    // pathObj.fillCol may be a function that generates dynamic color (so call it)
    if (pathObj.strokeCol instanceof Function)
    {
      this.ctx.fillStyle = pathObj.strokeCol(arguments);
    }
    else
    {
      this.ctx.fillStyle = pathObj.strokeCol;
    }
    // now actually fill the text
    this.ctx.fillText(pathObj.drawCmds, pathObj.imgX, pathObj.imgY);
    this.ctx.restore();   // undo the rotation
     // undo the translation and scaling
    this.ctx.restore();

    // construct the 'as rendered' outline (in pixels) for mouse event detection
    // if the text has been rotated set up some tools to rotate the pixel coords
    var A, sinA, cosA, rotXY;
    if (degs)
    {
      A = -Math.PI*degs/180.0;   // radians
      sinA = Math.sin(A);
      cosA = Math.cos(A);
      rotXY = function(x,y)
              {
                return [x*cosA - y*sinA, x*sinA + y*cosA];
              };
    }
    pathObj.pxOutline = [];   // start with new array
    // scale and translate to pixel coords (save these for drag and drop testing)
    pathObj.pxOrgX = this.vpLLx+this.xoffset+xPos*this.xscl;
    pathObj.pxOrgY = this.vpLLy+this.yoffset+yPos*this.yscl;
    // make a boundary path around the text
    var ulx = xScale*pathObj.imgX;
    var uly = xScale*pathObj.imgY;
    var urx = xScale*pathObj.imgX+xScale*pathObj.width;
    var ury = uly;
    var lrx = urx;
    var lry = xScale*pathObj.imgY+xScale*pathObj.height;
    var llx = ulx;
    var lly = lry;

    var vul = [ulx, uly];
    var vur = [urx, ury];
    var vlr = [lrx, lry];
    var vll = [llx, lly];
    // if the image has been rotated then rotate the bounding box
    if (degs)
    {
      vul = rotXY(ulx, uly);
      vur = rotXY(urx, ury);
      vlr = rotXY(lrx, lry);
      vll = rotXY(llx, lly);
    }
    // now form the DrawCmds and translate the vertices
    var moveXY = function(v)
                 {
                   return [pathObj.pxOrgX+v[0], pathObj.pxOrgY+v[1]];
                 };
    pathObj.pxOutline = [];   // start with new array
    pathObj.pxOutline[0] = new DrawCmd('moveTo', moveXY(vul));
    pathObj.pxOutline[1] = new DrawCmd('lineTo', moveXY(vur));
    pathObj.pxOutline[2] = new DrawCmd('lineTo', moveXY(vlr));
    pathObj.pxOutline[3] = new DrawCmd('lineTo', moveXY(vll));
    pathObj.pxOutline[4] = new DrawCmd('closePath');

    if (pathObj.dragNdrop != null)
    {
      // save pixel coords of the drawing origin (often useful in drag n drop)
      pathObj.dragNdrop.xOrg = xPos;
      pathObj.dragNdrop.yOrg = yPos;
      // now push it into Cango.draggable array, its checked by canvas mousedown event handler
      if (!_draggable[this.cId].contains(pathObj))
      {
        _draggable[this.cId].push(pathObj);
      }
    }
  };

  Cango.prototype.compilePath = function(path, color, scl, lineWidth, isotropic, drag)
  {
    /* this is a preprocessor to convert an array of svg Path syntax letters
       and numbers into 'segs' format, it treats a simple set of numbers as a
       special case, with an initial 'M' then the rest as 'L' pairs
       segs = [ ['M', c, c], ['L', c, c, c, c ..], [], []... ];
      which are then compiled to canvas drawCmd objects ready to render */
    if (!isArray(path))
    {
      return;
    }

    var iso = (isotropic === 'iso');
    var scale = scl || 1;
    var segs = [];
    var i, j;
    // special case of data only array. Test if 1st element is a number then
    // treat array as 'M', x0, y0, 'L', x1, y1, x2, y2, ... ]
    if (typeof path[0] == 'number')
    {
      segs[0] = ['M', path[0], path[1]];
      var lineSeg = ['L'];
      for (j=2,i=2; i<path.length; i++)
      {
        if (typeof path[i] != 'number')
        {
          break;
        }
      }
      segs[1] = lineSeg.concat(path.slice(j,i));
    }
    else
    {
      for(j=0, i=1; i<path.length; i++)
      {
        if (typeof path[i] == 'string')
        {
          segs.push(path.slice(j,i));
          j = i;
        }
      }
      segs.push(path.slice(j,i));    // push the last command out
    }
    // now send these off to the svg segs to canvas DrawCmd processor
    var cvsCmds = this.segsToDrawCmds(segs, 0, 0, scale, scale);
    var pathObj = new Obj2D(this, cvsCmds, "PATH", iso, null, color, null, drag);
    pathObj.strokeWidth = lineWidth || this.penWid;

    return pathObj;
  };

  Cango.prototype.compileShape = function(path, fillColor, strokeColor, scl, isotropic, drag)
  {
    // this is a preprocessor to convert an array of svg Path syntax letters and numbers into 'segs' format
    // segs = [ ['M', c, c], ['L', c, c, c, c ..], [], []... ];
    // which are then compiled to canvas drawCmd objects ready to render
    if (!isArray(path))
    {
      return;
    }

    var iso = (isotropic === 'iso');
    var scale = scl || 1;
    var segs = [];
    var i, j;
    // special case of data only array. Test if 1st element is a number then
    // treat array as 'M', x0, y0, 'L', x1, y1, x2, y2, ... ]
    if (typeof path[0] == 'number')
    {
      segs[0] = ['M', path[0], path[1]];
      var lineSeg = ['L'];
      for (j=2,i=2; i<path.length; i++)
      {
        if (typeof path[i] != 'number')
        {
          break;
        }
      }
      segs[1] = lineSeg.concat(path.slice(j,i));
    }
    else
    {
      for(j=0, i=1; i<path.length; i++)
      {
        if (typeof path[i] == 'string')
        {
          segs.push(path.slice(j,i));
          j = i;
        }
      }
      segs.push(path.slice(j,i));    // push the last command out
    }
    // now send these off to the svg segs-to-canvas DrawCmd processor
    var cvsCmds = this.segsToDrawCmds(segs, 0, 0, scale, scale);
    var pathObj = new Obj2D(this, cvsCmds, "SHAPE", iso, fillColor, strokeColor, null, drag);

    return pathObj;
  };

  Cango.prototype.compileText = function(str, color, fontSz, fontWt, lorigin, drag)
  {
    if (typeof str != 'string')
    {
      return;
    }
    var weight = this.fontWeight;   // default = 400
    if (typeof fontWt == 'string')
    {
      weight = fontWt;           // 'bold' etc
    }
    else if (isNumber(fontWt) && (fontWt > 99) && (fontWt < 901))
    {
      weight = fontWt;           // 100 .. 900
    }
    var size = fontSz || this.fontSize;     // fontSize in pts
    size *= 1.3;    // convert to pixels
    var lorg = lorigin || 1;

    var pathObj = new Obj2D(this, str, "TEXT", true, null, color, lorg, drag);
    this.ctx.save();
    // set the drawing context to measure the size
    this.ctx.font = size+"px "+this.fontFamily;
    var wid = this.ctx.measureText(str).width;   // need these for dragNdrop
    this.ctx.restore();
    var hgt = size;   // Note: char cell is ~1.5* fontSize pixels high
    var wid2 = wid/2;
    var hgt2 = hgt/2;
    var lorgWC = [0, [0, 0],    [wid2, 0],   [wid, 0],
                     [0, hgt2], [wid2, hgt2], [wid, hgt2],
                     [0, hgt],  [wid2, hgt],  [wid, hgt]];

    pathObj.imgX = -lorgWC[lorg][0];  // world coords offset to drawing origin
    pathObj.imgY = -lorgWC[lorg][1];
    pathObj.width = wid;
    pathObj.height = size;
    pathObj.fontSize = size;
    pathObj.fontWeight = weight;

    return pathObj;
  };

  Cango.prototype.compileImg = function(imgURL, w, lorigin, drag)
  {
    if (typeof imgURL != 'string')
    {
      return;
    }

    var lorg = lorigin || 1;
    var img = new Image();
    var imgObj = new Obj2D(this, img, "IMG", true, null, null, lorg, drag);   // iso=true, colors=null

    function onloadCallback(obj, w)
    {
      return function()
            {
              var wid = w || obj.drawCmds.width;    // its loaded now so width is valid
              var hgt = wid*obj.drawCmds.height/obj.drawCmds.width;  // keep aspect ratio
              var wid2 = wid/2;
              var hgt2 = hgt/2;
              var lorgWC = [0, [0, 0],    [wid2, 0],   [wid, 0],
                               [0, hgt2], [wid2, hgt2], [wid, hgt2],
                               [0, hgt],  [wid2, hgt],  [wid, hgt]];
              obj.imgX = -lorgWC[lorg][0];     // world coords offset to drawing origin
              obj.imgY = -lorgWC[lorg][1];
              obj.width = wid;
              obj.height = hgt;
            };
    }
    imgObj.drawCmds.src = "";
    addLoadEvent(imgObj.drawCmds, onloadCallback(imgObj, w));
    // start to load the image
    imgObj.drawCmds.src = imgURL;

    return imgObj;
  };

  Cango.prototype.compileSvgPath = function(svgPath, color, xRef, yRef, drag)
  {
    if (typeof svgPath != 'string')
    {
      return;
    }
    // this is a preprocessor to get an svg Path string into 'Obj2D' format
    var segs = [];
    var cmd, seg, cmdLetters, coords;
    var strs = svgPath.split(/(?=[a-df-z])/i);  // avoid e in exponents
    // now array of strings with command letter start to each
    var i;
    for (i=0; i<strs.length; i++)
    {
      seg = strs[i];
      // get command letter into an array
      cmdLetters = seg.match(/[a-z]/i);
      if (!cmdLetters)
      {
        return [];
      }
      cmd = cmdLetters.slice(0,1);
      if ((i==0)&&(cmd[0] != 'M'))   // check that the first move is absolute
      {
        cmd[0] = 'M';
      }
      coords = seg.match(/[\-\+]?[0-9]*\.?[0-9]+([eE][\-\+]?[0-9]+)?/gi);
      if (coords)
      {
        coords = coords.map(parseFloat);
      }
      segs.push(cmd.concat(coords));
    }
    // now send these off to the svg segs to canvas DrawCmd processor
    var xScale = 1;
    var yScale = -1;           // flip all the y coords to +ve up
    var xOfs = xRef || 0;      // move the drawing origin point
    var yOfs = yRef || 0;

    var drawCommands = this.segsToDrawCmds(segs, xOfs, -yOfs, xScale, yScale);
    // svg path are iso=true
    var pathObj = new Obj2D(this, drawCommands, "PATH", true, null, color, null, drag);

    return pathObj;
  };

  Cango.prototype.compileSvgShape = function(svgPath, fillColor, strokeColor, xRef, yRef, drag)
  {
    if (typeof svgPath != 'string')
    {
      return;
    }
    // this is a preprocessor to get an svg Path string into 'Obj2D' format
    var segs = [];
    var cmd, seg, cmdLetters, coords;
    var strs = svgPath.split(/(?=[a-df-z])/i);  // avoid e in exponents
    // now array of strings with command letter start to each
    var i;
    for (i=0; i<strs.length; i++)
    {
      seg = strs[i];
      // get command letter into an array
      cmdLetters = seg.match(/[a-z]/i);
      if (!cmdLetters)
      {
        return [];
      }
      cmd = cmdLetters.slice(0,1);
      if ((i==0)&&(cmd[0] != 'M'))   // check that the first move is absolute
      {
        cmd[0] = 'M';
      }
      coords = seg.match(/[\-\+]?[0-9]*\.?[0-9]+([eE][\-\+]?[0-9]+)?/gi);
      if (coords)
      {
        coords = coords.map(parseFloat);
      }
      segs.push(cmd.concat(coords));
    }
    // now send these off to the svg segs to canvas DrawCmd processor
    var xScale = 1;
    var yScale = -1;                        // flip all the y coords to +ve up
    var xOfs = xRef || 0;                 // move the shape reference point
    var yOfs = yRef || 0;
    var drawCommands = this.segsToDrawCmds(segs, xOfs, -yOfs, xScale, yScale);

    var pathObj = new Obj2D(this, drawCommands, "SHAPE", true, fillColor, strokeColor, null, drag);  // iso = true

    return pathObj;
  };

  Cango.prototype.drawPath = function(path, x, y, color, scl, isotropic)
  {
    var pathObj = this.compilePath(path, color, scl, null, isotropic);
    this.render(pathObj, x, y);
    return pathObj;
  };

  Cango.prototype.drawShape = function(path, x, y, fillColor, scl, isotropic)
  {
    // outline the same as fill color
    var pathObj = this.compileShape(path, fillColor, fillColor, scl, isotropic);
    this.render(pathObj, x, y);
    return pathObj;
  };

  Cango.prototype.drawText = function(str, x, y, color, ptSize, lorg)
  {
    var pathObj = this.compileText(str, color, ptSize, null, lorg);
    this.render(pathObj, x, y);
    return pathObj;
  };

  Cango.prototype.drawImg = function(imgURL, x, y, w, lorigin)  // just load img then calls _paintImg
  {
    var savThis = this;

    var xOfs = x || 0;
    var yOfs = y || 0;
    var lorg = lorigin || 1;
    var wid = w || 0;
    var img = new Image();
    var imgObj = new Obj2D(this, img, "IMG", true, null, null, lorg);   // iso = true for IMG

    function onloadCallback(obj, x, y, w)
    {
      return function()
              {
                var width = w || obj.drawCmds.width;    // its loaded now so width should be valid
                // only drawImg or compileImg can set width, render must wait till its loaded by one of these.
                var wid = width;   // world coords (save in object properties)
                var hgt = width*obj.drawCmds.height/obj.drawCmds.width;
                var wid2 = wid/2;
                var hgt2 = hgt/2;
                var lorgWC = [0, [0, 0],    [wid2, 0],   [wid, 0],
                                 [0, hgt2], [wid2, hgt2], [wid, hgt2],
                                 [0, hgt],  [wid2, hgt],  [wid, hgt]];
                obj.imgX = -lorgWC[lorg][0];     // world coords offset to drawing origin
                obj.imgY = -lorgWC[lorg][1];
                obj.width = wid;                 // world coords (save in object properties)
                obj.height = hgt;
                // now render the image, scale is set by width value (in world coords), 0 degs rotation
                // must use _paintImg not render to get the correct rotation values
                savThis._paintImg(obj, x, y, 1, 1, 0);  // scl = 1, degs = 0
              };
    }
    imgObj.drawCmds.src = "";
    // don't render image until dimensions are know (for scaling etc)
    addLoadEvent(imgObj.drawCmds, onloadCallback(imgObj, xOfs, yOfs, wid));
    // start to load the image
    imgObj.drawCmds.src = imgURL;

    return imgObj;
  };

  Cango.prototype.clipPath = function(pathObj)
  {
    if ((this.type == "IMG")||(this.type == "TEXT"))
    {
      return;
    }
    var pxlCoords = [];
    var i, j;
    this.ctx.save();             // not required when resetClip is supported
    this.ctx.beginPath();
    for (i=0; i<pathObj.drawCmds.length; i++)
    {
      for (j=0; j<pathObj.drawCmds[i].parms.length; j+=2)  // convert x,y coord pairs to pixel coords
      {
        pxlCoords[j] = this.vpLLx+this.xoffset+this.xscl*pathObj.drawCmds[i].parms[j];
        if (pathObj.iso)
        {
          pxlCoords[j+1] = this.vpLLy+this.yoffset-this.xscl*pathObj.drawCmds[i].parms[j+1];
        }
        else
        {
          pxlCoords[j+1] = this.vpLLy+this.yoffset+this.yscl*pathObj.drawCmds[i].parms[j+1];
        }
      }
      this.ctx[pathObj.drawCmds[i].drawFn].apply(this.ctx, pxlCoords); // actually draw the path
    }
    this.ctx.clip();
  };

  Cango.prototype.resetClip = function()
  {
    // this.ctx.resetClip();       // not supported in bowsers yet (Feb13)
    this.ctx.restore();            // use this until resetClip is supported
  };

  function interpolate(vals, t, dur)     // vals is an array of key frame values
  {
    if (isArray(vals))  // array
    {
      if (t >= dur)
      {
        return vals[vals.length-1];  // freeze at end value
      }
      var numSlabs = vals.length-1;
      var slabDur = dur/numSlabs;
      var slab = Math.floor(t/slabDur);
      var frac = (t - slab*slabDur)/slabDur;
      return vals[slab] + frac*(vals[slab+1] - vals[slab]);
    }
    //  else single value (not animated)
    return vals;
  }

  function DeclAnimation(obj, xValues, yValues, sclValues, rotValues, delayTime, duration, repeat)
  {
    if (xValues === undefined)
    {
      xValues = null;
    }
    if (yValues === undefined)
    {
      yValues = null;
    }
    if (sclValues === undefined)
    {
      sclValues = null;
    }
    if (rotValues === undefined)
    {
      rotValues = null;
    }

    this.delay = delayTime || 0;
    this.dur = duration || 0;     // 0 = go forever
    this.loop = repeat || false;
    this.startTime = 0;
    this.obj = obj;

    this.getState = function(time)
    {
      var state = {x:0, y:0, scl:0, rot:0};
      if (time == 0)   // re-starting after a stop
      {
        this.startTime = 0;
      }
      var localTime = time - this.startTime;       // handles local looping
      if ((localTime > this.dur+this.delay) && (this.dur > 0) && this.loop)  // dur = 0 means go forever
      {
        this.startTime = time;   // we will re-start
        localTime = 0;      // this will force re-start at frame 0
      }
      var t = 0;
      if (localTime > this.delay)  // repeat initial frame if there is a delay to start
      {
        t = localTime - this.delay;
      }
      if (isArray(xValues))
      {
        state.x = interpolate(xValues, t, this.dur);
      }
      else
      {
        state.x = xValues;      // single value or null
      }
      if (isArray(yValues))
      {
        state.y = interpolate(yValues, t, this.dur);
      }
      else
      {
        state.y = yValues;      // single value or null
      }
      if (isArray(sclValues))
      {
        state.scl = interpolate(sclValues, t, this.dur);
      }
      else
      {
        state.scl = sclValues;  // single value or null
      }
      if (isArray(rotValues))
      {
        state.rot = interpolate(rotValues, t, this.dur);
      }
      else
      {
        state.rot = rotValues;  // single value or null
      }

      return state;
    };
  }

  Cango.prototype.animate = function(obj, xValues, yValues, sclValues, rotValues, delayTime, duration, loopStr)
  {
    var delay = delayTime || 0;
    var dur = duration || 0;     // 0 = go forever
    var loop = false;
    if ((loopStr != undefined)&&(loopStr == 'loop'))
    {
      loop = true;
    }

    var newAnim = new DeclAnimation(obj, xValues, yValues, sclValues, rotValues, delay, dur, loop);
    // give it a unique ID
    newAnim.id = this.cId+"_"+this.uniqueId();
    // push this into the Cango animations array
    this.animations.push(newAnim);

    return newAnim.id;   // so the animation just created can be deleted if required
  };

  Cango.prototype.stopAnimation = function()
  {
    window.cancelAnimationFrame(this.timer);
    this.prevAnimMode = this.animMode;
    this.animMode = this.modes.STOPPED;
    // reset the currTime so play and step know to start again
    this.currTime = 0;
  };

  Cango.prototype.pauseAnimation = function()
  {
    window.cancelAnimationFrame(this.timer);
    this.prevAnimMode = this.animMode;
    this.animMode = this.modes.PAUSED;
  };

  Cango.prototype.stepAnimation = function()
  {
    var savThis = this;

    // make the buffered drawing context match the screen ctx
    if (this.buffered)
    {
      this.bufCtx.strokeStyle = this.penCol;
      this.bufCtx.fillStyle = this.paintCol;
      this.bufCtx.lineWidth = this.penWid;
      this.bufCtx.lineCap = this.lineCap;
    }
    // eqivalent to play for one frame and pause
    if (this.animMode == this.modes.PLAYING)
    {
      return;
    }
    if (this.animMode == this.modes.PAUSED)
    {
      this.startTime = Date.now() - this.currTime;  // move time as if currFrame just drawn
    }
    this.prevAnimMode = this.animMode;
    this.animMode = this.modes.STEPPING;

    // this is the actual animator that draws the frame
    function drawIt()
    {
      var nextState = null;
      var A = null;
      var time = Date.now();    // use this as a time stamp, browser don't all pass the same time code
      if (savThis.prevAnimMode == savThis.modes.STOPPED)
      {
        savThis.startTime = time;                // forces localTime = 0 to start from beginning
      }

      var localTime =  time - savThis.startTime;
      // do all the drawing for this frame
      savThis.clearCanvas();
      var temp;
      if (savThis.buffered)
      {
        // drawing will be off screen, clear buffer too
        savThis.bufCtx.clearRect(0, 0, savThis.rawWidth, savThis.rawHeight);
        // swap buffers while drawing done off scrreen
        temp = savThis.ctx;
        savThis.ctx = savThis.bufCtx;
      }
      var i;
      for (i=0; i<savThis.animations.length; i++)
      {
        A = savThis.animations[i];       // is an animation object with getState method
        nextState = A.getState(localTime);
        savThis.render(A.obj, nextState.x, nextState.y, nextState.scl, nextState.rot);
      }
      if (savThis.buffered)
      {
        // drawing done, switch them back
        savThis.ctx = temp;
        // now bit-blt the image in buffer to the on-screen canvas
        savThis.ctx.drawImage(_buf[savThis.cId], 0, 0);
      }

      savThis.currTime = localTime;      // timestamp of what is currently on screen
      savThis.prevAnimMode = savThis.modes.PAUSED;
      savThis.animMode = savThis.modes.PAUSED;
      //
    }

    setTimeout(drawIt, this.stepTime);
  };

  Cango.prototype.playAnimation = function()
  {
    // anim is an object with id, obj, getState etc properties
    // each object in the animations are rendered even if anim.mode = stopped
    // since the canvas is cleared every frame and the stopped ones would disappear
    // to have them disappear use deleteAnimation()
    // all aniamtions get passed the same value of local time each frame to keep sync
    // this routine is the 'stepper' frim timeline
    var savThis = this;

    // make the buffered drawing context match the screen ctx
    if (this.buffered)
    {
      this.bufCtx.strokeStyle = this.penCol;
      this.bufCtx.fillStyle = this.paintCol;
      this.bufCtx.lineWidth = this.penWid;
      this.bufCtx.lineCap = this.lineCap;
    }
    if (this.animMode == this.modes.PLAYING)
    {
      return;
    }
    if (this.animMode == this.modes.PAUSED)
    {
      this.startTime = Date.now() - this.currTime;  // move time as if currFrame just drawn
    }

    this.prevAnimMode = this.animMode;
    this.animMode = this.modes.PLAYING;

    // this is the actual animator that draws each frame
    function drawIt()
    {
      var nextState = null;
      var A = null;
      var time = Date.now();    // use this as a time stamp, browser don't all pass the same time code
      if (savThis.prevAnimMode == savThis.modes.STOPPED)
      {
        savThis.startTime = time;                // forces localTime = 0 to start from beginning
      }

      var localTime =  time - savThis.startTime;
      // do all the drawing for this frame
      savThis.clearCanvas();
      var temp;
      if (savThis.buffered)
      {
        // drawing will be off screen, clear buffer too
        savThis.bufCtx.clearRect(0, 0, savThis.rawWidth, savThis.rawHeight);
        // swap buffers while drawing done off scrreen
        temp = savThis.ctx;
        savThis.ctx = savThis.bufCtx;
      }
      var i;
      for (i=0; i<savThis.animations.length; i++)
      {
        A = savThis.animations[i];       // is an animation object with getState method
        nextState = A.getState(localTime);
        savThis.render(A.obj, nextState.x, nextState.y, nextState.scl, nextState.rot);
      }
      if (savThis.buffered)
      {
        // drawing done, switch them back
        savThis.ctx = temp;
        // now bit-blt the image in buffer to the on-screen canvas
        savThis.ctx.drawImage(_buf[savThis.cId], 0, 0);
      }
      // drawing done
      savThis.currTime = localTime;      // timestamp of what is currently on screen
      savThis.prevAnimMode = savThis.modes.PLAYING;
      savThis.timer = window.requestAnimationFrame(drawIt);
    }

    this.timer = window.requestAnimationFrame(drawIt);
  };

  Cango.prototype.deleteAnimation = function(animId)
  {
    this.stopAnimation();
    var idx = -1;
    var i;
    for (i=0; i<this.animations.length; i++)
    {
      if (this.animations[i] == animId)
      {
        idx = i;
        break;
      }
    }
    if (idx == -1)
    {
      // not found
      return;
    }
    this.animations.splice(idx,1);       // delete the animation object
  };

  Cango.prototype.deleteAllAnimations = function()
  {
    this.stopAnimation();
    this.animations = [];
  };

  Cango.prototype.segsToDrawCmds = function (segs, xRef, yRef, xScl, yScl)
  {
    var x = 0;
    var y = 0;
    var px, py;
    var c1x, c1y;
    var rot, rx, ry, larc, swp;
    var arc_segs;
    var cmd, pc;
    var cmdObj;
    var seg, coords;
    var commands = [];
    var xScale = xScl || 1;
    var yScale = yScl || 1;
    var xOfs = xRef || 0;          // move the shape drawing origin
    var yOfs = yRef || 0;
    var i, j;
    for (i=0; i<segs.length; i++)
    {
      seg = segs[i];
      cmd = seg[0];
      if ((i==0)&&(cmd != 'M'))   // check that the first move is absolute
      {
        cmd = 'M';
      }
      coords = seg.slice(1);      // skip the command copy coords
      if (coords)
      {
        coords = coords.map(parseFloat);
      }
      switch(cmd)
      {
        case 'M':
          x = xOfs + xScale*coords[0];
          y = yOfs + yScale*coords[1];
          px = py = null;
          cmdObj = new DrawCmd('moveTo', [x, y]);
          commands.push(cmdObj);
          coords.splice(0, 2);      // delete the 2 coords from the front of the array
          while (coords.length>0)
          {
            x = xOfs + xScale*coords[0];                // eqiv to muliple 'L' calls
            y = yOfs + yScale*coords[1];
            cmdObj = new DrawCmd('lineTo', [x, y]); // any coord pair after first move is regarded as line
            commands.push(cmdObj);
            coords.splice(0, 2);
          }
          break;
        case 'm':
          x += xScale*coords[0];
          y += yScale*coords[1];
          px = py = null;
          cmdObj = new DrawCmd('moveTo', [x, y]);
          commands.push(cmdObj);
          coords.splice(0, 2);      // delete the 2 coords from the front of the array
          while (coords.length>0)
          {
            x += xScale*coords[0];              // eqiv to muliple 'l' calls
            y += yScale*coords[1];
            cmdObj = new DrawCmd('lineTo', [x, y]); // any coord pair after first move is regarded as line
            commands.push(cmdObj);
            coords.splice(0, 2);
          }
          break;
        case 'L':
          while (coords.length>0)
          {
            x = xOfs + xScale*coords[0];
            y = yOfs + yScale*coords[1];
            cmdObj = new DrawCmd('lineTo', [x, y]);
            commands.push(cmdObj);
            coords.splice(0, 2);
          }
          px = py = null;
          break;
        case 'l':
          while (coords.length>0)
          {
            x += xScale*coords[0];
            y += yScale*coords[1];
            cmdObj = new DrawCmd('lineTo', [x, y]);
            commands.push(cmdObj);
            coords.splice(0, 2);
          }
          px = py = null;
          break;
        case 'H':
          x = xOfs + xScale*coords[0];
          px = py = null ;
          cmdObj = new DrawCmd('lineTo', [x, y]);
          commands.push(cmdObj);
          break;
        case 'h':
          x += xScale*coords[0];
          px = py = null ;
          cmdObj = new DrawCmd('lineTo', [x, y]);
          commands.push(cmdObj);
          break;
        case 'V':
          y = yOfs + yScale*coords[0];
          px = py = null;
          cmdObj = new DrawCmd('lineTo', [x, y]);
          commands.push(cmdObj);
          break;
        case 'v':
          y += yScale*coords[0];
          px = py = null;
          cmdObj = new DrawCmd('lineTo', [x, y]);
          commands.push(cmdObj);
          break;
        case 'C':
          while (coords.length>0)
          {
            c1x = xOfs + xScale*coords[0];
            c1y = yOfs + yScale*coords[1];
            px = xOfs + xScale*coords[2];
            py = yOfs + yScale*coords[3];
            x = xOfs + xScale*coords[4];
            y = yOfs + yScale*coords[5];
            cmdObj = new DrawCmd('bezierCurveTo', [c1x, c1y, px, py, x, y]);
            commands.push(cmdObj);
            coords.splice(0, 6);
          }
          break;
        case 'c':
          while (coords.length>0)
          {
            c1x = x + xScale*coords[0];
            c1y = y + yScale*coords[1];
            px = x + xScale*coords[2];
            py = y + yScale*coords[3];
            x += xScale*coords[4];
            y += yScale*coords[5];
            cmdObj = new DrawCmd('bezierCurveTo', [c1x, c1y, px, py, x, y]);
            commands.push(cmdObj);
            coords.splice(0, 6);
          }
          break;
        case 'S':
          if (px == null || !pc.match(/[sc]/i))
          {
            px = x;                // already absolute coords
            py = y;
          }
          cmdObj = new DrawCmd('bezierCurveTo', [x-(px-x), y-(py-y),
                                                xOfs + xScale*coords[0], yOfs + yScale*coords[1],
                                                xOfs + xScale*coords[2], yOfs + yScale*coords[3]]);
          commands.push(cmdObj);
          px = xOfs + xScale*coords[0];
          py = yOfs + yScale*coords[1];
          x = xOfs + xScale*coords[2];
          y = yOfs + yScale*coords[3];
          break;
        case 's':
          if (px == null || !pc.match(/[sc]/i))
          {
            px = x;
            py = y;
          }
          cmdObj = new DrawCmd('bezierCurveTo', [x-(px-x), y-(py-y),
                                                x + xOfs + xScale*coords[0], y + yOfs + yScale*coords[1],
                                                x + xOfs + xScale*coords[2], y + yOfs + yScale*coords[3]]);
          commands.push(cmdObj);
          px = x + xScale*coords[0];
          py = y + yScale*coords[1];
          x += xScale*coords[2];
          y += yScale*coords[3];
          break;
        case 'Q':
          px = xOfs + xScale*coords[0];
          py = yOfs + yScale*coords[1];
          x = xOfs + xScale*coords[2];
          y = yOfs + yScale*coords[3];
          cmdObj = new DrawCmd('quadraticCurveTo', [px, py, x, y]);
          commands.push(cmdObj);
          break;
        case 'q':
          cmdObj = new DrawCmd('quadraticCurveTo', [x + xScale*coords[0], y + yScale*coords[1],
                                                    x + xScale*coords[2], y + yScale*coords[3]]);
          commands.push(cmdObj);
          px = x + xScale*coords[0];
          py = y + yScale*coords[1];
          x += xScale*coords[2];
          y += yScale*coords[3];
          break;
        case 'T':
          if ((px == null) || (!pc.match(/[qt]/i)))
          {
            px = x;
            py = y;
          }
          else
          {
            px = x-(px-x);
            py = y-(py-y);
          }
          cmdObj = new DrawCmd('quadraticCurveTo', [px, py,
                                                    xOfs + xScale*coords[0], yOfs + yScale*coords[1]]);
          commands.push(cmdObj);
          px = x-(px-x);
          py = y-(py-y);
          x = xOfs + xScale*coords[0];
          y = yOfs + yScale*coords[1];
          break;
        case 't':
          if (px == null || !pc.match(/[qt]/i))
          {
            px = x;
            py = y;
          }
          else
          {
            px = x-(px-x);
            py = y-(py-y);
          }
          cmdObj = new DrawCmd('quadraticCurveTo', [px, py,
                                                    x + xScale*coords[0], y + yScale*coords[1]]);
          commands.push(cmdObj);
          x += xScale*coords[0];
          y += yScale*coords[1];
          break;
        case 'A':
          while (coords.length>0)
          {
            px = x;
            py = y;
            rx = xScale*coords[0];
            ry = xScale*coords[1];
            rot = -coords[2];          // rotationX: swap for CCW +ve
            larc = coords[3];          // large arc    should be ok
            swp = 1 - coords[4];       // sweep: swap for CCW +ve
            x = xOfs + xScale*coords[5];
            y = yOfs + yScale*coords[6];
            arc_segs = this.arcToBezier(px, py, rx, ry, rot, larc, swp, x, y);
            for (j=0; j<arc_segs.length; j++)
            {
              cmdObj = new DrawCmd('bezierCurveTo', arc_segs[j]);
              commands.push(cmdObj);
            }
            coords.splice(0, 7);
          }
          break;
        case 'a':
          while (coords.length>0)
          {
            px = x;
            py = y;
            rx = xScale*coords[0];
            ry = xScale*coords[1];
            rot = -coords[2];          // rotationX: swap for CCW +ve
            larc = coords[3];          // large arc    should be ok
            swp = 1 - coords[4];       // sweep: swap for CCW +ve
            x += xScale*coords[5];
            y += yScale*coords[6];
            arc_segs = this.arcToBezier(px, py, rx, ry, rot, larc, swp, x, y);
            for (j=0; j<arc_segs.length; j++)
            {
              cmdObj = new DrawCmd('bezierCurveTo', arc_segs[j]);
              commands.push(cmdObj);
            }
            coords.splice(0, 7);
          }
          break;
        case 'Z':
          cmdObj = new DrawCmd('closePath', []);
          commands.push(cmdObj);
          break;
        case 'z':
          cmdObj = new DrawCmd('closePath', []);
          commands.push(cmdObj);
          break;
      }
      pc = cmd;     // save the previous command for possible reflected control points
    }
    return commands;
  };

  Cango.prototype.arcToBezier = function(ox, oy, rx, ry, rotateX, large, sweep, x, y)
  {
    var th = rotateX * (Math.PI/180);
    var sin_th = Math.sin(th);
    var cos_th = Math.cos(th);
    rx = Math.abs(rx);
    ry = Math.abs(ry);
    var px = cos_th * (ox - x) * 0.5 + sin_th * (oy - y) * 0.5;
    var py = cos_th * (oy - y) * 0.5 - sin_th * (ox - x) * 0.5;
    var pl = (px*px) / (rx*rx) + (py*py) / (ry*ry);
    if (pl > 1)
    {
      pl = Math.sqrt(pl);
      rx *= pl;
      ry *= pl;
    }
    var a00 = cos_th / rx;
    var a01 = sin_th / rx;
    var a10 = -sin_th / ry;
    var a11 = cos_th / ry;
    var x0 = a00 * ox + a01 * oy;
    var y0 = a10 * ox + a11 * oy;
    var x1 = a00 * x + a01 * y;
    var y1 = a10 * x + a11 * y;
    var d = (x1-x0) * (x1-x0) + (y1-y0) * (y1-y0);
    var sfactor_sq = 1 / d - 0.25;
    if (sfactor_sq < 0)
    {
      sfactor_sq = 0;
    }
    var sfactor = Math.sqrt(sfactor_sq);
    if (sweep == large)
    {
      sfactor = -sfactor;
    }
    var xc = 0.5 * (x0 + x1) - sfactor * (y1-y0);
    var yc = 0.5 * (y0 + y1) + sfactor * (x1-x0);

    var th0 = Math.atan2(y0-yc, x0-xc);
    var th1 = Math.atan2(y1-yc, x1-xc);

    var th_arc = th1-th0;
    if (th_arc < 0 && sweep == 1)
    {
      th_arc += 2*Math.PI;
    }
    else if (th_arc > 0 && sweep == 0)
    {
      th_arc -= 2 * Math.PI;
    }
    var segments = Math.ceil(Math.abs(th_arc / (Math.PI * 0.5 + 0.001)));
    var result = [];
    var i, th2, th3;
    for (i=0; i<segments; i++)
    {
      th2 = th0 + i * th_arc / segments;
      th3 = th0 + (i+1) * th_arc / segments;
      result.push(this.segmentToBezier(xc, yc, th2, th3, rx, ry, sin_th, cos_th));
    }

    return result;
  };

  Cango.prototype.segmentToBezier = function(cx, cy, th0, th1, rx, ry, sin_th, cos_th)
  {
    var a00 = cos_th * rx;
    var a01 = -sin_th * ry;
    var a10 = sin_th * rx;
    var a11 = cos_th * ry;

    var th_half = 0.5 * (th1 - th0);
    var t = (8/3) * Math.sin(th_half * 0.5) * Math.sin(th_half * 0.5) / Math.sin(th_half);
    var x1 = cx + Math.cos(th0) - t * Math.sin(th0);
    var y1 = cy + Math.sin(th0) + t * Math.cos(th0);
    var x3 = cx + Math.cos(th1);
    var y3 = cy + Math.sin(th1);
    var x2 = x3 + t * Math.sin(th1);
    var y2 = y3 - t * Math.cos(th1);
    return [
              a00 * x1 + a01 * y1, a10 * x1 + a11 * y1,
              a00 * x2 + a01 * y2, a10 * x2 + a11 * y2,
              a00 * x3 + a01 * y3, a10 * x3 + a11 * y3
            ];
  };

  // copy the basic graphics context values (for an overlay)
  Cango.prototype.dupCtx = function(src_graphCtx)
  {
    // copy all the graphics context parameters into the overlay ctx.
    this.vpLLx = src_graphCtx.vpLLx;      // vp lower left from canvas left in pixels
    this.vpLLy = src_graphCtx.vpLLy;      // vp lower left from canvas top
    this.xscl = src_graphCtx.xscl;        // world x axis scale factor
    this.yscl = src_graphCtx.yscl;        // world y axis scale factor
    this.xoffset = src_graphCtx.xoffset;  // world x origin offset from viewport left in pixels
    this.yoffset = src_graphCtx.yoffset;  // world y origin offset from viewport bottom in pixels
    this.penCol = src_graphCtx.penCol.slice(0);   // copy value not reference
    this.penWid = src_graphCtx.penWid;    // pixels
    this.lineCap = src_graphCtx.lineCap.slice(0);
    this.paintCol = src_graphCtx.paintCol.slice(0);
    this.fontFamily = src_graphCtx.fontFamily.slice(0);
    this.fontSize = src_graphCtx.fontSize;
    this.fontWeight = src_graphCtx.fontWeight;
  };

  function Drag2D(cangoGC, grabFn, dragFn, dropFn)
  {
    var savThis = this;

    this.cgo = cangoGC;
    this.parent = null;
    this.grabCallback = grabFn || null;
    this.dragCallback = dragFn || null;
    this.dropCallback = dropFn || null;
    this.xOrg = 0;      // drawing origin (world coords), updated by render or enableDrag
    this.yOrg = 0;
    this.xGrabOfs = 0;  // csr offset (world coords) from origin filled in at grab time
    this.yGrabOfs = 0;

    // these closures are called in the scope of the Drag2D instance so this is valid
    this.grab = function(evt)
    {
      var event = evt||window.event;
      this.cgo.cnvs.onmouseup = function(event)
                                {
                                  savThis.drop(event);
                                };
      var csrPos = this.cgo.getCursorPos(event);      // update mouse pos to pass to the owner
      var csrPosWC = this.cgo.toWorldCoords(csrPos.x, csrPos.y);
      // save the cursor offset from the drawing origin
      this.xGrabOfs = csrPosWC.x - this.xOrg;
      this.yGrabOfs = csrPosWC.y - this.yOrg;

      if (this.grabCallback)
      {
        this.grabCallback(csrPosWC);    // call in the scope of dragNdrop object
      }

      this.cgo.cnvs.onmousemove = function(event)
                                  {
                                    savThis.drag(event);
                                  };
      if (event.preventDefault)       // prevent default browser action (W3C)
      {
        event.preventDefault();
      }
      else                        // shortcut for stopping the browser action in IE
      {
        window.event.returnValue = false;
      }
      return false;
    };

    this.drag = function(event)
    {
      var csrPos = this.cgo.getCursorPos(event);  // update mouse pos to pass to the owner
      var csrPosWC = this.cgo.toWorldCoords(csrPos.x, csrPos.y);
      if (this.dragCallback)
      {
        this.dragCallback(csrPosWC);
      }
    };

    this.drop = function(event)
    {
      var csrPos = this.cgo.getCursorPos(event);  // update mouse pos to pass to the owner
      var csrPosWC = this.cgo.toWorldCoords(csrPos.x, csrPos.y);
      this.cgo.cnvs.onmouseup = null;
      this.cgo.cnvs.onmousemove = null;
      if (this.dropCallback)
      {
        this.dropCallback(csrPosWC);
      }
    };
  }

