//-------------------------------------------------------------------
// Helper functions for remote scripting using IFRAMES (originally)
// Inspired by developer.apple.com/internet/webcontent/iframe.html .
// Later (in 2013) extended by 'XMLHttpRequest' which - HERE - 
//    doesn't have anything to do with XML because SL doesn't support
//    XML (too bulky, slow, complex, unnnecessary, .. ).
//    but to read JSON-formatted data from Spectrum Lab .
//-------------------------------------------------------------------

var IFrameObj; // our IFrame object (old fashioned, doesn't depend on XMLHttpRequest)

var xmlHttp=null; // one instance of an XMLHttpRequest (doesn't work on stoneage browsers)

function CallServer(url)  // calls the server in SpecLab through an IFRAME 
{                         // (old school; NOT used for JSON - see ReadViaHttpRequest)
  if (!document.createElement) {return true};
  var IFrameDoc;
  // Encode the expression into the part of a URI:
  if (!IFrameObj && document.createElement) 
   {
    // create the IFrame and assign a reference to the
    // object to our global variable IFrameObj.
    // this will only happen the first time 
    // callToServer() is called
   try {
      var tempIFrame=document.createElement('iframe');
      tempIFrame.setAttribute('id','RSIFrame');
      tempIFrame.style.border='0px';
      tempIFrame.style.width='0px';
      tempIFrame.style.height='0px';
      IFrameObj = document.body.appendChild(tempIFrame);
      
      if (document.frames) {
        // this is for IE5 Mac, because it will only
        // allow access to the document object
        // of the IFrame if we access it through
        // the document.frames array
        IFrameObj = document.frames['RSIFrame'];
      }
    } catch(exception) {
      // This is for IE5 PC, which does not allow dynamic creation
      // and manipulation of an iframe object. Instead, we'll fake
      // it up by creating our own objects.
      iframeHTML='\<iframe id="RSIFrame" style="';
      iframeHTML+='border:0px;';
      iframeHTML+='width:0px;';
      iframeHTML+='height:0px;';
      iframeHTML+='"><\/iframe>';
      document.body.innerHTML+=iframeHTML;
      IFrameObj = new Object();
      IFrameObj.document = new Object();
      IFrameObj.document.location = new Object();
      IFrameObj.document.location.iframe = document.getElementById('RSIFrame');
      IFrameObj.document.location.replace = function(location) {
        this.iframe.src = location;
      }
    }
  }
  
  if (navigator.userAgent.indexOf('Gecko') !=-1 && !IFrameObj.contentDocument) {
    // we have to give NS6 a fraction of a second
    // to recognize the new IFrame
    setTimeout('callToServer()',10);
    return false;
  }
  
  if (IFrameObj.contentDocument) {
    // For NS6
    IFrameDoc = IFrameObj.contentDocument; 
  } else if (IFrameObj.contentWindow) {
    // For IE5.5 and IE6
    IFrameDoc = IFrameObj.contentWindow.document;
  } else if (IFrameObj.document) {
    // For IE5
    IFrameDoc = IFrameObj.document;
  } else {
    return true;
  }
  
  IFrameDoc.location.replace(url);
  return false;
}


function CallIFunc(key,expression)  // calls a single INTERPRETER FUNCTION in SpecLab
{
  // Encode the expression into the part of a URI:
  return CallServer('_ifunc.html?'+encodeURIComponent(key)+'='+encodeURIComponent( expression ) );
  // The result will be passed to the application via HandleIFunc(key,result),
  // as soon as the remote web server has sent the response .
}


function CallIProc(commandline)  // execute a line with interpreter commands in SpecLab
{
  // Encode the command into the part of a URI:
  return CallServer('_iproc.html?'+encodeURIComponent( commandline ) );
  // The result will be passed to the application via HandleIProc(error_string) .
  // as soon as the remote web server has sent the response .
}

function ReadViaHttpRequest( filename_or_url )
{
  if( xmlHttp==null ) // there's no 'XML'-HTTP-Request-thing yet (WHAT a name..)
   {                  // .. so try to create one
     try 
      { // Mozilla, Opera, Safari and Internet Explorer (since v7)
        xmlHttp = new XMLHttpRequest();
      } 
     catch(e) 
      {
        try 
         { // MS Internet Explorer (ab v6)
           xmlHttp  = new ActiveXObject("Microsoft.XMLHTTP");
         } 
        catch(e) 
         {
           try 
            { // MS Internet Explorer (ab v5)
              xmlHttp  = new ActiveXObject("Msxml2.XMLHTTP");
            } 
           catch(e) 
            {
              xmlHttp  = null;
            }
         }
      }
   }   
  if (xmlHttp)  // successfully created this "XML"-HTTP-Request-thing, so use it:
   {
     xmlHttp.open('GET', filename_or_url, true);
     // Note: Trying this locally (without a server, but with the HTML loaded FROM A FILE)
     // didn't work with Chrome/Iron: "Cross origin requests are only supported for HTTP" .
     // Holy shit. To debug this stuff, we MUST run it on a server (like the one in SL) !
     xmlHttp.onreadystatechange = function () 
       {
         if (xmlHttp.readyState == 4) // 4 indicates "DONE" ('the data transfer has been completed or something went wrong')
          {
            // alert(xmlHttp.responseText);
          }
       };
     xmlHttp.send(null);
     return xmlHttp.responseText;
   }
  else  // no joy with the XML-HTTP-Request-thingy
   { return null;
   }   
} // end ReadViaHttpRequest()

